function fns = ConstructRFLBROCKProb_St(D, N, fns)
% D is the dataset, stroed in cell.
% N is number of sample
% fns is a structure related to Problem, and contains 
%    fns.cost, 
%    fns.grad,
%    fns.partialgrad, 

    fns.cost  = @(x) cost(x, D);
    fns.grad = @(x) partialgrad(x, D, 1:N);
    fns.partialgrad = @(x, S) partialgrad(x, D, S);    
end

function output = cost(x, D)
    Asumx = sum(D.A, 3)' * x;
    Nx = D.N .* x;
    output = sum(sum(Asumx .* Nx)) / size(D.A, 3);

    % tmp = 0;
    % for i = 1:N
    %     tmp = tmp + singlecost(x, D{i});
    % end
    % output = tmp / N;
end

function output = partialgrad(x, D, S)
    As = D.A(:,:,S);
    Nx  = (D.N .* x);            
    tmp = (2/length(S)) * sum(As, 3) * Nx;
    
    sym = x'*tmp;
    sym = 0.5 *(sym + sym');
    output = tmp - x*sym;

    % tmp = zeros(size(x));
    % for i = reshape(S, 1, length(S))
    %     tmp = tmp + singleegrad(x, D{i});
    % end
    % output = tmp / length(S);
end

function output = singlecost(x, Di)
    tmp1 = x' * Di.A;
    tmp2 = diag(Di.N)' .* x; % X * N
    output = trace(tmp1 * tmp2);
end

function output = singleegrad(x, Di)
    output = 2 * Di.A * (diag(Di.N)' .* x);
end



