function fns = ConstructRFLFMProb_SPD(D, N, fns)
% D is the dataset, stroed in cell.
% N is number of sample
% fns is a structure related to Problem, and contains 
%    fns.cost, 
%    fns.egrad,
%    fns.partialegrad, 

    fns.cost  = @(x) cost(x, D, N);
    fns.egrad = @(x) partialegrad(x, D, 1:N);
    fns.partialegrad = @(x, S) partialegrad(x, D, S);
    
    fns.grad = @(x) partialgrad(x, D, 1:N);
    fns.partialgrad = @(x, S) partialgrad(x, D, S);

%     fns.grad = @(x) fns.M.egrad2rgrad(x, fns.egrad(x));
%     fns.partialgrad = @(x, S) fns.M.egrad2rgrad(x, fns.partialegrad(x, S));
    
    symm = @(x) 0.5 *(x + x');
    vec = @(A) A(:);
    trAB = @(A, B) vec(A')' * vec(B); % trace(A*B)
    trAA = @(A) (trAB(A, A));         % (trace(A^2))
    
    function output = cost(x, D, N)
        tmp = 0;
        for i = 1:N
            tmp = tmp + singlecost(x, D{i});
        end
        output = tmp / N;
    end
    
    % partial Euclidean gradient
    function output = partialegrad(x, D, S)
        tmp = zeros(size(x));
        for i = reshape(S, 1, length(S))
            tmp = tmp + singleegrad(x, D{i});
        end
        output = tmp / length(S);
    end
    
    % partial Riemannian gradient
    function output = partialgrad(x, D, S)
        tmp = zeros(size(x));
        for i = reshape(S, 1, length(S))
            tmp = tmp + singlegrad(x, D{i});
        end
        output = tmp / length(S);
    end    
    
    function output = singlecost(x, Di)
        sm = sqrtm(x);
        tmp = logm(sm \ Di / sm);
        output = trAA(real(tmp));
    end
    
    % Euclidean gradient for f_ij
    function output = singleegrad(x, Di)
        sm = sqrtm(x);    
        output = (-2) * symm(sm \ real(logm(sm \ Di / sm)) / sm);
    end

    % Riemannian gradient for f_ij
    function output = singlegrad(x,  Di)
        output = - 2 * fns.M.log(x, Di);
    end

end