set(0,'defaultaxesfontsize',22, ...
   'defaultaxeslinewidth',0.7, ...
   'defaultlinelinewidth',.8,'defaultpatchlinewidth',0.8);
set(0,'defaultlinemarkersize',10)

clear;close all;clc
rng(0);

d = 25;
p = 2;
Manifold = stiefelfactory(d, p);
Manifold.transp = @isotransp;

nagents = 20;
ncostterms = 50;
K1 = 1;
%%% construct agents

prob = cell(nagents,1);
opts = cell(nagents,1);

D = cell(nagents,1);
for i = 1:nagents      
    A = zeros(d, d, ncostterms);
    for j = 1:ncostterms
        AA = randn(d, d);
        n = 1 + (p-1) * rand(p,1);        
        A(:,:,j) = AA + AA';   
    end
    D{i}.A = A;
    D{i}.N = p:-1:1;

    probtmp.M = Manifold;    
    probtmp.ncostterms = ncostterms;
    probtmp = ConstructRFLBROCKProb_St(D{i}, ncostterms, probtmp);
    prob{i} = probtmp;
               
    optstmp.batchsize = ceil(probtmp.ncostterms * 0.5) ; 
    optstmp.verbosity = 0;
    optstmp.checkperiod = 1;
    optstmp.maxiter = K1;
    opts{i} = optstmp;       
    % checkgradient(prob{i})
end

problem.M = Manifold;
problem.nagents = nagents;
problem.agent = prob;
problem.cost = @(x) cost(problem, x);
problem.grad = @(x) grad(problem, x);
options.verbosity = 1;

options.checkperiod = 1;
options.agent = opts;
options.maxiter = 800;
options.localMethod = 'RSGD';
options.option = 1;
options.decaysteps = 50;

rng(1)
x0 = Manifold.rand();

% for step size
options.stepsize_type = 'decay';
% options.stepsize_type = 'fix';
options.stepsize = .02;
options.s = nagents;

Seed = 10;

fvs1 = [];
ngfs1 = [];
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K1)
    [x1, fv1, info1] = RFedAGS(problem, x0, options);
    fvs1 = [fvs1, info1.fvs];
    ngfs1 = [ngfs1, info1.ngfs];
end
% figure(1)
% t = 0:options.checkperiod:options.maxiter;
% plot(t,mean(fvs1,2), '-^', 'LineWidth',2)

%%%%%% K = 4 %%%%%%%
fvs2 = [];
ngfs2 = [];
K2 = 4;
for i = 1:nagents
    options.agent{i}.maxiter = K2;
end
for seed = 1:Seed
    fprintf('seed:%d, K:%d\n', seed, K2)
%     rng(seed)
    [x2, fv2, info2] = RFedAGS(problem, x0, options);
    fvs2 = [fvs2, info2.fvs];
    ngfs2 = [ngfs2, info2.ngfs];
end
% figure(1)
% hold on
% plot(t,mean(fvs2,2), '->', 'LineWidth',2)

%%%%%% K = 8 %%%%%%%
fvs3 = [];
ngfs3 = [];
K3 = 8;
for i = 1:nagents
    options.agent{i}.maxiter = K3;
end
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K3)
    [x3, fv3, info3] = RFedAGS(problem, x0, options);
    fvs3 = [fvs3, info3.fvs];
    ngfs3 = [ngfs3, info3.ngfs];
end
% figure(1)
% hold on
% plot(t,mean(fvs3,2), '-v', 'LineWidth',2)

%%%%%% K = 12 %%%%%%%
fvs4 = [];
ngfs4 = [];
K4 = 12;
for i = 1:nagents
    options.agent{i}.maxiter = K4;
end
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K4)
    [x4, fv4, info4] = RFedAGS(problem, x0, options);
    fvs4 = [fvs4, info4.fvs];
    ngfs4 = [ngfs4, info4.ngfs];
end
% figure(1)
% hold on
% plot(t,mean(fvs4,2), '-<', 'LineWidth',2)

%%%%%% K = 14 %%%%%%%
fvs5 = [];
ngfs5 = [];
K5 = 14;
for i = 1:nagents
    options.agent{i}.maxiter = K5;
end
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K5)
    [x5, fv5, info5] = RFedAGS(problem, x0, options);
    fvs5 = [fvs5, info5.fvs];
    ngfs5 = [ngfs5, info5.ngfs];
end
% figure(1)
% hold on
% plot(t,mean(fvs5,2), '-o', 'LineWidth',2)

problem.M = stiefelfactory(d, p);
[x6, fv6, info6] = rlbfgs(problem, x0);

% figure(1)
% hold on
% plot(0:length([info6.cost])-1,[info6.cost], 'LineWidth',2)
% xlabel('Iteration','FontSize', 18)
% ylabel('Cost','FontSize', 18)
% legend('RFedAGS-K='+string(K1), 'RFedAGS-K='+string(K2), 'RFedAGS-K='+string(K3),...
%        'RFedAGS-K='+string(K4), 'RFedAGS-K='+string(K5), 'RSD', 'FontSize', 18)
% 
% saveas(gca, 'BROCK_Syn_K_Decaying'+string(nagents)+'x'+string(ncostterms) + '_Cost')
% saveas(gca, 'BROCK_Syn_K_Decaying'+string(nagents)+'x'+string(ncostterms) + '_Cost.eps')

fvs1 = mean(fvs1,2);
fvs2 = mean(fvs2,2);
fvs3 = mean(fvs3,2);
fvs4 = mean(fvs4,2);
fvs5 = mean(fvs5,2);

save('BROCK_Syn_K_Decaying'+string(nagents)+'x'+string(ncostterms)+'.mat', "nagents", "ncostterms", "fvs1","fvs2","fvs3","fvs4", "fvs5", "fv6")

figure(2)
t = 0:options.checkperiod:options.maxiter;
semilogy(t, mean(fvs1,2)-fv6, '-^', t, mean(fvs2,2)-fv6, '->', t, mean(fvs3,2)-fv6, '-v',...
         t, mean(fvs4,2)-fv6, '-<', t, mean(fvs5,2)-fv6, '-o', 'LineWidth', 2)
xlabel('Iteration','FontSize', 18)
ylabel('Excess Risk','FontSize', 18)
legend('RFedAGS-K='+string(K1), 'RFedAGS-K='+string(K2), 'RFedAGS-K='+string(K3),...
       'RFedAGS-K='+string(K4), 'RFedAGS-K='+string(K5), 'FontSize', 18)

saveas(gca, 'BROCK_Syn_K_Decaying'+string(nagents)+'x'+string(ncostterms) + '_ER')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function out = cost(problem, x)
    out = problem.agent{1}.cost(x);
    for i = 2:problem.nagents
        out = out + problem.agent{i}.cost(x);
    end
    out = out / problem.nagents;
end

function out = grad(problem, x)
    out = problem.agent{1}.grad(x);
    for i = 2:problem.nagents
        out = out + problem.agent{i}.grad(x);
    end
    out = out / problem.nagents;
end

%%%%%%%% isometric vector transport by parallelization
function out = isotransp(x, y, v)    
    global ybases;   
    if isempty(ybases)
        ybases = build_bases_embedded(y);
    end
    xbases = build_bases_embedded(x);
    [n, p] = size(v);    
    [U, ~, V] = svd(full(ybases' * xbases));
    v = reshape(v, n * p, 1);
    out = reshape(ybases * (U * (V' * (xbases' * v))), n, p);
end

function output = build_bases_embedded(x)
    [n, p] = size(x);
    x_perp = null(x');
    x_perp = x_perp(1 : n, 1 : n - p);
    x = x / sqrt(2);
    d = n * p - p * (p + 1) / 2;
    output = sparse(n * p, d);
    col = 1;
    for i = 2 : p
        x_i_1 = x(:, i - 1);
        for j = i : p
            output((i - 2) * n + 1 : (i - 1) * n, col) = - x(:, j);
            output((j - 1) * n + 1 : j * n, col) = x_i_1;
            col = col + 1;
        end
    end
    for i = 1 : p
        output((i - 1) * n + 1 : i * n, col : col + n - p - 1) = x_perp;
        col = col + n - p;
    end
end

