set(0,'defaultaxesfontsize',22, ...
   'defaultaxeslinewidth',0.7, ...
   'defaultlinelinewidth',.8,'defaultpatchlinewidth',0.8);
set(0,'defaultlinemarkersize',10)

clear;close all;clc
rng(0);

d = 25 - 1;
Manifold = spherefactory(d+1, 1);
Manifold.type = 'sphere';
Manifold.ManProj = @(x) ManProj(x);

nagents = 10;
ncostterms = 80;
K1 = 5;
%%% construct agents

prob = cell(nagents,1);
opts = cell(nagents,1);

v = 1e-3;
D = zeros(ncostterms * nagents, d+1);
for i = 1:nagents    
    tmp = [1.0, 1.0 - 1.1*v, 1.0 - 1.2*v, 1.0 - 1.3*v, 1.0-1.4*v, randn(1, d+1 - 5) / (d+1)]; 
    Sigma = diag(tmp);  
    U = orth(randn(ncostterms, d+1));
    V = orth(randn(d+1));
    Z = V'*Sigma*U';
    D((i-1)*ncostterms+1:i*ncostterms,:) = Z';

    probtmp.M = Manifold;
    probtmp.ncostterms = ncostterms;
    probtmp = ConstructRFLPCAProb_Stief(D((i-1)*ncostterms+1:i*ncostterms,:), ncostterms, probtmp);
    prob{i} = probtmp;           
    
    optstmp.batchsize = ceil(probtmp.ncostterms * 0.8) ; 
    optstmp.verbosity = 0;
    optstmp.checkperiod = 1;
    optstmp.maxiter = K1;    
    opts{i} = optstmp;       
%     checkgradient(prob{i})
end

problem.M = Manifold;
problem.nagents = nagents;
problem.agent = prob;
problem.cost = @(x) cost(problem, x, D);
problem.grad = @(x) grad(problem, x, D);

options.checkperiod = 1;
options.agent = opts;
options.maxiter = 800;
options.localMethod = 'RSGD';
options.option = 1;
options.decaysteps = 30;
options.lambda = 0.1;
options.verbosity = 1;

rng(1)
x0 = Manifold.rand();

% for step size
options.stepsize_type = 'decay';
% options.stepsize_type = 'fix';
options.stepsize = 1;
options.s = nagents;

Seed = 10;

fvs1 = [];
ngfs1 = [];
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K1)
    [x1, fv1, info1] = RFedAGS(problem, x0, options);
    fvs1 = [fvs1, info1.fvs];
    ngfs1 = [ngfs1, info1.ngfs];
end
% figure(1)
% t = 0:options.checkperiod:options.maxiter;
% semilogy(t,mean(fvs1,2), 'LineWidth',2)

%%%%%% K = 5 %%%%%%%
fvs2 = [];
ngfs2 = [];
K2 = 10;
for i = 1:nagents
    options.agent{i}.maxiter = K2;
end
for seed = 1:Seed
    fprintf('seed:%d, K:%d\n', seed, K2)
%     rng(seed)
    [x2, fv2, info2] = RFedAGS(problem, x0, options);
    fvs2 = [fvs2, info2.fvs];
    ngfs2 = [ngfs2, info2.ngfs];
end
% figure(1)
% hold on
% semilogy(t,mean(fvs2,2), 'LineWidth',2)

%%%%%% K = 10 %%%%%%%
fvs3 = [];
ngfs3 = [];
K3 = 20;
for i = 1:nagents
    options.agent{i}.maxiter = K3;
end
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K3)
    [x3, fv3, info3] = RFedAGS(problem, x0, options);
    fvs3 = [fvs3, info3.fvs];
    ngfs3 = [ngfs3, info3.ngfs];
end
% figure(1)
% hold on
% semilogy(t,mean(fvs3,2), 'LineWidth',2)

%%%%%% K = 20 %%%%%%%
fvs4 = [];
ngfs4 = [];
K4 = 30;
for i = 1:nagents
    options.agent{i}.maxiter = K4;
end
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K4)
    [x4, fv4, info4] = RFedAGS(problem, x0, options);
    fvs4 = [fvs4, info4.fvs];
    ngfs4 = [ngfs4, info4.ngfs];
end
% figure(1)
% hold on
% semilogy(t,mean(fvs4,2), 'LineWidth',2)

[x5, fv5, info5] = steepestdescent(problem, x0);

% figure(1)
% hold on
% semilogy(0:length([info5.cost])-1,[info5.cost], 'LineWidth',2)
% xlabel('Iteration','FontSize', 18)
% ylabel('Cost','FontSize', 18)
% legend('RFedAGS-K='+string(K1), 'RFedAGS-K='+string(K2), 'RFedAGS-K='+string(K3), 'RFedAGS-K='+string(K4), 'RSD', 'FontSize', 18)
% 
% saveas(gca, 'PCAST_Syn_K_Decaying'+string(nagents)+'x'+string(ncostterms) + '_Cost')
% saveas(gca, 'PCAST_Syn_K_Decaying'+string(nagents)+'x'+string(ncostterms) + '_Cost.eps','psc')

fvs1 = mean(fvs1,2);
fvs2 = mean(fvs2,2);
fvs3 = mean(fvs3,2);
fvs4 = mean(fvs4,2);

save('PCAST_Syn_K_Decaying'+string(nagents)+'x'+string(ncostterms)+'.mat',"nagents", 'ncostterms',"fvs1","fvs2","fvs3","fvs4", "fv5")

figure(2)
t = 0:options.checkperiod:options.maxiter;
semilogy(t, mean(fvs1,2)-fv5, '-^', t, mean(fvs2,2)-fv5, '->',... 
         t, mean(fvs3,2)-fv5, '-v', t, mean(fvs4,2)-fv5, '-<', 'LineWidth', 2)
xlabel('Iteration','FontSize', 18)
ylabel('Excess Risk','FontSize', 18)
legend('RFedAGS-K='+string(K1), 'RFedAGS-K='+string(K2), 'RFedAGS-K='+string(K3), 'RFedAGS-K='+string(K4), 'FontSize', 18)

saveas(gca, 'PCAST_Syn_K_Decaying'+string(nagents)+'x'+string(ncostterms) + '_ER')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function out = cost(problem, x, D)
    Dx = D * x;
    out = - sum(sum(Dx .* Dx)) / size(D,1);
end

function out = grad(problem, x, D)
    tmp = (-2/size(D,1)) * D' * (D * x);
    sym = x'*tmp;
    sym = 0.5 * (sym+ sym');
    out= tmp - x * sym;
end












