close all;clear;clc;
rng(0);

d = 2;
Manifold = sympositivedefinitefactory(d);
Manifold.type = 'SPD';
Manifold.transp=@Manifold.paralleltransp;
Manifold.retr = @Manifold.exp;
Manifold.invretr = @Manifold.log;

Dw = 1;

numSamples = 600;

nagents = 20;
ncostterms = ceil(numSamples / nagents);

fprintf('numSamples:%d, nagents:%d, ncostterms:%d\n', numSamples, nagents, ncostterms)

Z = cell(numSamples, 1);
t = 0;
while t < numSamples
    tmp = wishrnd(eye(d)/ 2, 2);
%        tmp = rand(2,2);
    if dist(tmp, eye(d)) <= Dw
       Z{t+1} = tmp;
       t = t+1;
    end    
end

K1 = 1;
for i = 1:nagents

    probtmp.M = Manifold;
    probtmp.ncostterms = ncostterms;
    D = {Z{(i-1)*ncostterms+1 : i*ncostterms}};
    probtmp = ConstructRFLFMProb_SPD(D, ncostterms, probtmp);
    prob{i} = probtmp;

    metrics.cost = @(problem, x) getCost(problem, x);
    metrics.gradnorm = @(problem, x) problem.M.norm(x, getGradient(problem, x));
    optstmp.statsfun = statsfunhelper(metrics); 
    optstmp.batchsize = ceil(probtmp.ncostterms * .5);
    optstmp.verbosity = 1;
    optstmp.checkperiod = 1;
    optstmp.maxiter = K1; 
    opts{i} = optstmp;
%     checkgradient(prob{i})
end

problem.M = Manifold;
problem.nagents = nagents;
problem.agent = prob;
problem.cost = @(x) cost(problem, x);
problem.grad = @(x) grad(problem, x);

options.checkperiod = 1;
options.agent = opts;
options.maxiter = 500;
options.localMethod = 'RSGD';
options.option = 1;
options.verbosity = 0;
options.decaysteps = 10;

rng(1)
x0 = Manifold.rand();

% for step size
% options.stepsize_type = 'decay';
options.stepsize_type = 'fix';
options.stepsize = .009;
options.s = nagents;

Seed = 10;

fvs1  = [];
ngfs1 = [];
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K1)
    [x1, fv1, info1] = RFedAGS(problem, x0, options);
    fvs1 = [fvs1, info1.fvs];
    ngfs1 = [ngfs1, info1.ngfs];
end
% figure(1)
% t = 0:options.checkperiod:options.maxiter;
% semilogy(t,mean(fvs1,2), 'LineWidth',2)

%%%%%% K = 4 %%%%%%%
fvs2 = [];
ngfs2 = [];
K2 = 2;
for i = 1:nagents
    options.agent{i}.maxiter = K2;
end
for seed = 1:Seed
    fprintf('seed:%d, K:%d\n', seed, K2)
%     rng(seed)
    [x2, fv2, info2] = RFedAGS(problem, x0, options);
    fvs2 = [fvs2, info2.fvs];
    ngfs2 = [ngfs2, info2.ngfs];
end
% figure(1)
% hold on
% semilogy(t,mean(fvs2,2), 'LineWidth',2)

%%%%%% K = 8 %%%%%%%
% fvs3 = [];
% ngfs3 = [];
% K3 = 3;
% for i = 1:nagents
%     options.agent{i}.maxiter = K3;
% end
% for seed = 1:Seed
% %     rng(seed)
%     fprintf('seed:%d, K:%d\n', seed, K3)
%     [x3, fv3, info3] = RFedAGS(problem, x0, options);
%     fvs3 = [fvs3, info3.fvs];
%     ngfs3 = [ngfs3, info3.ngfs];
% end
% figure(1)
% hold on
% semilogy(t,mean(fvs3,2), 'LineWidth',2)

%%%%%% K = 10 %%%%%%%
% fvs4 = [];
% ngfs4 = [];
% K4 = 10;
% for i = 1:nagents
%     options.agent{i}.maxiter = K4;
% end
% for seed = 1:Seed
% %     rng(seed)
%     fprintf('seed:%d, K:%d\n', seed, K4)
%     [x4, fv4, info4] = RFedAvg(problem, x0, options);
%     fvs4 = [fvs4, info4.fvs];
%     ngfs4 = [ngfs4, info4.ngfs];
% end
% figure(1)
% hold on
% semilogy(t,mean(fvs4,2), 'LineWidth',2)

[x5, fv5, info5] = conjugategradient(problem, x0);

% figure(1)
% hold on
% semilogy(0:length([info5.cost])-1,[info5.cost], 'LineWidth',2)
% xlabel('Iteration','FontSize', 18)
% ylabel('Cost','FontSize', 18)
% legend('RFedAGS-K='+string(K1), 'RFedAGS-K='+string(K2), 'RFedAGS-K='+string(K3), 'RFedAGS-K='+string(K4), 'RSD', 'FontSize', 18)

% saveas(gca, 'FMSPD_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_Cost')
% saveas(gca, 'FMSPD_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_Cost.eps')


% figure(2)
% semilogy(t, mean(fvs1,2)-fv5, '-^', t, mean(fvs2,2)-fv5, '->',...
%          t, mean(fvs3,2)-fv5, '-v', t, mean(fvs4,2)-fv5, '-<', 'LineWidth', 2)
% xlabel('Iteration','FontSize', 18)
% ylabel('Excess Risk','FontSize', 18)
% legend('RFedAGS-K='+string(K1), 'RFedAGS-K='+string(K2), 'RFedAGS-K='+string(K3), 'RFedAGS-K='+string(K4), 'FontSize', 18)
% 
% 
% saveas(gca, 'FMSPD_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_ER')
% saveas(gca, 'FMSPD_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_ER.eps')


fvs1 = mean(fvs1,2);
fvs2 = mean(fvs2,2);
% fvs3 = mean(fvs3,2);
% fvs4 = mean(fvs4,2);
% 
% save('FMSPD_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms)+'.mat',"fvs1","fvs2","fvs3","fvs4", "fv5")

opt.legends = ["RFedAGS-K="+string(K1), "RFedAGS-K="+string(K2)];
opt.ylabel = 'Excess Risk';
opt.Markers = {"-^", "->", "-v", "-<"};
t = 0:options.checkperiod:options.maxiter;
opt.MarkerIndices = [20,20,20];
opt.Markersize = 8;
opt.LineWidth = 2.1;
opt.number = 1;
opt.Color4 = 0.8;
opt.xlabel = 'Iterations';
opt.filePath = 'FMSPD_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_12';
myplot([t', t'], [fvs1, fvs2] - fv5, opt)



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function out = dist(x, y)
    sm = sqrtm(x);
    tmp = real(logm(sm \ y / sm));
    out = sqrt( sum(sum(tmp .* tmp)) );
end

function out = cost(problem, x)
    tmp = 0;
    for i = 1:problem.nagents
        tmp = tmp + problem.agent{i}.cost(x);
    end
    out = tmp / problem.nagents;
end

function out = grad(problem, x)
    out = 0;
    for i = 1:problem.nagents
        out = out + problem.agent{i}.grad(x);
    end
    out = out / problem.nagents;
end


