close all;clear;clc;
rng(0);

d = 2;
Manifold = sympositivedefinitefactory(d);
Manifold.type = 'SPD';
Manifold.transp=@Manifold.paralleltransp;
Manifold.retr = @Manifold.exp;
Manifold.invretr = @Manifold.log;

Dw = 1;

numSamples = 5000;

nagents = 50;
ncostterms = ceil(numSamples / nagents);

fprintf('numSamples:%d, nagents:%d, ncostterms:%d\n', numSamples, nagents, ncostterms)

Z = cell(numSamples, 1);
t = 0;
while t < numSamples
    tmp = wishrnd(eye(d)/ 2, 2);
%        tmp = rand(2,2);
    if dist(tmp, eye(d)) <= Dw
       Z{t+1} = tmp;
       t = t+1;
    end    
end

K1 = 10;
for i = 1:nagents

    probtmp.M = Manifold;
    probtmp.ncostterms = ncostterms;
    D = {Z{(i-1)*ncostterms+1 : i*ncostterms}};
    probtmp = ConstructRFLFMProb_SPD(D, ncostterms, probtmp);
    prob{i} = probtmp;
    
    optstmp.batchsize = ceil(probtmp.ncostterms * .5);    
    optstmp.maxiter = K1; 
    opts{i} = optstmp;
%     checkgradient(prob{i})
end

problem.M = Manifold;
problem.nagents = nagents;
problem.agent = prob;
problem.cost = @(x) cost(problem, x);
problem.grad = @(x) grad(problem, x);

options.checkperiod = 1;
options.agent = opts;
options.maxiter = 200;
options.option = 1;
options.decaysteps = 10;
options.verbosity = 1;

rng(1)
x0 = Manifold.rand();

[x5, fv5, info5] = conjugategradient(problem, x0);


% for step size
% options.stepsize_type = 'decay';
options.stepsize_type = 'fix';
options.stepsize = .003;

Seed = 10;

fvs1  = [];
B1 = ceil(probtmp.ncostterms * 0.1);
for i = 1:nagents
    options.agent{i}.batchsize = B1;
end
for seed = 1:Seed
    fprintf('seed:%d, S:%d\n', seed, B1)
    [x1, fv1, info1] = RFedAGS(problem, x0, options);
    fvs1 = [fvs1, info1.fvs];    
end
figure(1)
t = 0:options.checkperiod:options.maxiter;
semilogy(t, mean(fvs1,2) - fv5, '-^', 'LineWidth',2)

%%%%%% K = 4 %%%%%%%
fvs2 = [];
% (S,B) = (30, 50)
B2 = ceil(probtmp.ncostterms * 0.4);
options.stepsize = .003 * sqrt(4/1);
for i = 1:nagents
    options.agent{i}.batchsize = B2;
end
for seed = 1:Seed
    fprintf('seed:%d, S:%d\n', seed, B2)
    [x2, fv2, info2] = RFedAGS(problem, x0, options);
    fvs2 = [fvs2, info2.fvs];    
end
figure(1)
hold on
semilogy(t,mean(fvs2,2)-fv5, '->', 'LineWidth',2)


%%%%%% K = 4 %%%%%%%
fvs3 = [];
% (S,B) = (30, 50)
B3 = ceil(probtmp.ncostterms * 0.8);
options.stepsize = .003 * sqrt(8/1);
for i = 1:nagents
    options.agent{i}.batchsize = B3;
end
for seed = 1:Seed
    fprintf('seed:%d, S:%d\n', seed, B3)
    [x3, fv3, info3] = RFedAGS(problem, x0, options);
    fvs3 = [fvs3, info3.fvs];    
end
figure(1)
hold on
semilogy(t,mean(fvs3,2)-fv5, '-v', 'LineWidth',2)

xlabel('Iteration','FontSize', 18)
ylabel('Excess Risk','FontSize', 18)
legend('RFedAGS-B='+string(B1), 'RFedAGS-B='+string(B2), 'RFedAGS-B='+string(B3), 'FontSize', 18)

saveas(gca, 'FMSPD_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_ER_B')

fvs1 = mean(fvs1,2);
fvs2 = mean(fvs2,2);
fvs3 = mean(fvs3,2);

save('FMSPD_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms)+'_B.mat',"nagents", 'ncostterms',"fvs1","fvs2","fvs3", "fv5")


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function out = dist(x, y)
    sm = sqrtm(x);
    tmp = real(logm(sm \ y / sm));
    out = sqrt( sum(sum(tmp .* tmp)) );
end

function out = cost(problem, x)
    tmp = 0;
    for i = 1:problem.nagents
        tmp = tmp + problem.agent{i}.cost(x);
    end
    out = tmp / problem.nagents;
end

function out = grad(problem, x)
    out = 0;
    for i = 1:problem.nagents
        out = out + problem.agent{i}.grad(x);
    end
    out = out / problem.nagents;
end


