set(0,'defaultaxesfontsize',22, ...
   'defaultaxeslinewidth',0.7, ...
   'defaultlinelinewidth',.8,'defaultpatchlinewidth',0.8);
set(0,'defaultlinemarkersize',10)

clear;close all;clc
rng(0);

d = 30-1;
p = 10;
Manifold = stiefelfactory(d+1, p);
Manifold.transp = @VTbyCayley;
Manifold.retr = @retr_cayley;
Manifold.invretr = @inv_retr_cayley;


nagents = 40;
ncostterms = 80;
%%% generate dataset
data = zeros(ncostterms * nagents, d+1);
v = 1e-3;
for i = 1:nagents    
    tmp = [1.0, 1.0 - 1.1*v, 1.0 - 1.2*v, 1.0 - 1.3*v, 1.0-1.4*v, randn(1, d+1 - 5) / (d+1)]; 
    Sigma = diag(tmp);  
    U = orth(randn(ncostterms, d+1));
    V = orth(randn(d+1));
    Z = V'*Sigma*U';
    data((i-1)*ncostterms+1:i*ncostterms,:) = Z';
end

rng(1)
x0 = Manifold.rand();

verbosity = 0;
stepsize = 0.8;
maxiter = 2000;
checkperiod = 1;
K1 = 10;

Seed = 10;

[problem, options] = generate_prob(data, Manifold, nagents, ncostterms, ceil(ncostterms * 0.5), K1);
options.stepsize_type = 'fix';
options.stepsize = stepsize;
options.checkperiod = checkperiod;
options.maxiter = maxiter;
options.tolgradnorm = 1e-8;
options.verbosity = verbosity;
[x5, fv5, info5] = rlbfgs(problem, x0, options);


fvs1 = [];
ngfs1 = [];
times1 = [];
for seed = 1:Seed
    fprintf('RFedAGS, Sseed:%d, K:%d\n', seed, K1) 
    [x1, fv1, info1] = RFedAGS(problem, x0, options); 
    fvs1 = [fvs1, info1.fvs];
    ngfs1 = [ngfs1, info1.ngfs];
    times1 = [times1, info1.CPUtime];
end
% figure(1)
% t = 0:options.checkperiod:options.maxiter;
% semilogy(t, mean(fvs1,2) - fv5, '-^','LineWidth',2)
% 
% figure(2)
% semilogy(mean(times1,2), mean(fvs1,2) - fv5,'-^', 'LineWidth',2)

%%%%%% RFedAvg  %%%%%%%
Manifold = stiefelfactory(d+1, p);
Manifold.transp = @VTbyCayley;
Manifold.retr = @retr_cayley;
Manifold.invretr = @inv_retr_cayley;

K2 = K1;
[problem, options] = generate_prob(data, Manifold, nagents, ncostterms, ceil(ncostterms * 0.5), K2);
options.stepsize_type = 'fix';
options.stepsize = stepsize;
options.checkperiod = checkperiod;
options.maxiter = maxiter;
options.verbosity = verbosity;
fvs2 = [];
ngfs2 = [];
times2 = [];
for seed = 1:Seed
    fprintf('RFedAvg, Sseed:%d, K:%d\n', seed, K2)    
    [x2, fv2, info2] = RFedAvg(problem, x0, options);
    fvs2 = [fvs2, info2.fvs];
    ngfs2 = [ngfs2, info2.ngfs];
    times2 = [times2, info2.CPUtime];
end
% figure(1)
% hold on
% semilogy(t,mean(fvs2,2) - fv5, '->', 'LineWidth',2)
% 
% figure(2)
% hold on
% semilogy(mean(times2,2), mean(fvs2,2) - fv5,'->', 'LineWidth',2)

%%%%%% RFedSVRG %%%%%%%
Manifold = stiefelfactory(d+1, p);
Manifold.transp = @VTbyCayley;
Manifold.retr = @retr_cayley;
Manifold.invretr = @inv_retr_cayley;

K3 = K1;
[problem, options] = generate_prob(data, Manifold, nagents, ncostterms, ceil(ncostterms * 0.5), K3);
options.stepsize_type = 'fix';
options.stepsize = stepsize;
options.checkperiod = checkperiod;
options.maxiter = maxiter;
options.verbosity = verbosity;
fvs3 = [];
ngfs3 = [];
times3 = [];
K3 = K1;
for i = 1:nagents
    options.agent{i}.maxiter = K3;
end
for seed = 1:Seed
    fprintf('RFedSVRG, Sseed:%d, K:%d\n', seed, K3) 
    [x3, fv3, info3] = RFedSVRG(problem, x0, options);
    fvs3 = [fvs3, info3.fvs];
    ngfs3 = [ngfs3, info3.ngfs];
    times3 = [times3, info3.CPUtime];
end
% figure(1)
% hold on
% semilogy(t,mean(fvs3,2) - fv5, '-v', 'LineWidth',2)
% 
% figure(2)
% hold on
% semilogy(mean(times3,2), mean(fvs3,2) - fv5, '-v', 'LineWidth',2)

% %%%%%% RFedProj %%%%%%%
Manifold = stiefelfactory(d+1, p);

K4 = K1;
[problem, options] = generate_prob(data, Manifold, nagents, ncostterms, ceil(ncostterms * 0.5), K4);
options.stepsize_type = 'fix';
options.stepsize = stepsize;
options.checkperiod = checkperiod;
options.maxiter = maxiter;
options.verbosity = verbosity;

fvs4 = [];
ngfs4 = [];
times4 = [];
for seed = 1:Seed
    fprintf('RFedProj, Sseed:%d, K:%d\n', seed, K4)    
    [x4, fv4, info4] = RFedProj(problem, x0, options);    
    fvs4 = [fvs4, info4.fvs];
    ngfs4 = [ngfs4, info4.ngfs];
    times4 = [times4, info4.CPUtime];
end
% figure(1)
% hold on
% semilogy(t,mean(fvs4,2) - fv5, '-<', 'LineWidth',2)
% title('K='+string(K1), 'FontSize', 18, 'Interpreter','latex')
% xlabel('Iteration','FontSize', 18)
% ylabel('Excess Risk','FontSize', 18)
% legend('RFedAGS', 'RFedAvg', 'RFedSVRG','RFedProj', 'FontSize', 18)
% 
% saveas(gca, 'PCASTIEF_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_cayley_ERvsIter')
% saveas(gca, 'PCASTIEF_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_cayley_ERvsIter.eps')
% 
% figure(2)
% hold on
% semilogy(mean(times4,2), mean(fvs4,2) - fv5, '-<', 'LineWidth',2)
% title('K='+string(K1), 'FontSize', 18, 'Interpreter','latex')
% xlabel('CPU time (s)','FontSize', 18)
% ylabel('Excess Risk','FontSize', 18)
% legend('RFedAGS', 'RFedAvg', 'RFedSVRG','RFedProj', 'FontSize', 18)
% 
% saveas(gca, 'PCASTIEF_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_cayley_ERvsTime')
% saveas(gca, 'PCASTIEF_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_cayley_ERvsTime.eps')

times1 = mean(times1, 2);
fvs1 = mean(fvs1,2);
times2 = mean(times2, 2);
fvs2 = mean(fvs2,2);
times3 = mean(times3, 2);
fvs3 = mean(fvs3,2);
times4 = mean(times4, 2);
fvs4 = mean(fvs4,2);

save('PCASTIEF_'+string(nagents)+'x'+string(ncostterms)+ '_' + string(p) +'_cayley.mat', "times1", "times2","times3","times4","fvs1","fvs2","fvs3","fvs4", "fv5")

opt.legends = ["RFedAGS", "RFedAvg", "RFedSVRG","RFedProj"];
opt.title = '$K$='+string(K1);
opt.ylabel = 'Excess Risk';
opt.Markers = {"-^", "->", "-v", "-<"};
t = 0:options.checkperiod:options.maxiter;
opt.MarkerIndices = [125,100,125,125];
opt.Markersize = 8;
opt.LineWidth = 2.1;
opt.number = 1;
opt.Color4 = 0.8;
opt.xlabel = 'Iterations';
opt.filePath = 'PCASTIEF_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_' + string(p) + '_cayley_ERvsIter';
myplot([t', t', t', t'], [fvs1, fvs2, fvs3, fvs4] - fv5, opt)

opt.xlabel = 'CPU time (s)';
opt.number = 2;
opt.filePath = 'PCASTIEF_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_' + string(p) +  '_cayley_ERvsTime';
myplot([times1, times2, times3, times4], [fvs1, fvs2, fvs3, fvs4] - fv5, opt)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [problem, options] = generate_prob(D, Manifold, nagents, ncostterms, batchsize, K)
    
    prob = cell(nagents, 1);
    opts = cell(nagents, 1);
    for i = 1:nagents

        probtmp.M = Manifold;
        probtmp.ncostterms = ncostterms;
        probtmp = ConstructRFLPCAProb_Stief(D((i-1)*ncostterms+1:i*ncostterms,:), ncostterms, probtmp);
        prob{i} = probtmp;

        optstmp.batchsize = batchsize;        
        optstmp.maxiter = K;        
        opts{i} = optstmp; 
%         checkgradient(probtmp)
    end
    problem.nagents = nagents;
    problem.agent = prob;
    problem.M = Manifold;
    problem.cost = @(x) cost(problem, x, D);
    problem.grad = @(x) grad(problem, x, D);

    options.agent = opts;
    options.option = 1;
end


function out = cost(problem, x, D)
    Dx = D * x;
    out = - sum(sum(Dx .* Dx)) / size(D,1);
end

function out = grad(problem, x, D)
    tmp = (-2/size(D,1)) * D' * (D * x);
    sym = x'*tmp;
    sym = 0.5 * (sym+ sym');
    out= tmp - x * sym;
end


%%%%%%%% isometric vector transport by parallelization
function out = isotransp(x, y, v)    
    global ybases;   
    if isempty(ybases)
        ybases = build_bases_embedded(y);
    end
    xbases = build_bases_embedded(x);
    [n, p] = size(v);    
    [U, ~, V] = svd(full(ybases' * xbases));
    v = reshape(v, n * p, 1);
    out = reshape(ybases * (U * (V' * (xbases' * v))), n, p);
end

function output = build_bases_embedded(x)
    [n, p] = size(x);
    x_perp = null(x');
    x_perp = x_perp(1 : n, 1 : n - p);
    x = x / sqrt(2);
    d = n * p - p * (p + 1) / 2;
    output = sparse(n * p, d);
    col = 1;
    for i = 2 : p
        x_i_1 = x(:, i - 1);
        for j = i : p
            output((i - 2) * n + 1 : (i - 1) * n, col) = - x(:, j);
            output((j - 1) * n + 1 : j * n, col) = x_i_1;
            col = col + 1;
        end
    end
    for i = 1 : p
        output((i - 1) * n + 1 : i * n, col : col + n - p - 1) = x_perp;
        col = col + n - p;
    end
end


% function out = retr_cayley(X, V, a)
%     [n, ~] = size(X);
%     if nargin == 3 && abs(a - 1) > 1e-16
%         V = a * V;
%     end
%     In = eye(n);
%     tmp = (V - 0.5 * X * (X' * V));
%     W = tmp * X'; 
%     W = 0.5 * (W - W');
%     Wl = 0.5 * [tmp, X];
%     Wr = [X, -tmp];
%     out = woodbury_inverse(Wl, Wr) * (X + W*X);
% end

function out = retr_cayley(X, Z, a)
    [~, p] = size(X);
    if nargin == 3 && abs(a - 1) > 1e-16
        Z = a * Z;
    end    
    XZ = X' * Z;
    [TMP, XZ5] = Auxiliary(X, XZ, Z);
    % out = X +  TMP(:, 1:p) - TMP(:,p+1:end) * XZ5';
    tmp = vertcat(eye(p), -XZ5);
    % out = X  + TMP * [eye(p); -XZ5];
    out = X + TMP * tmp;
end

function out = inv_retr_cayley(X, Y)
    %  Y = retr(X, out)
    [~, p] = size(X);
    tmp = (eye(p) + X' * Y); 
    out = 2 * (Y / tmp + X / (tmp') - X);
end

function out = VTbyCayley(X, Y, H)
    % Y = retr(X, Z);    
    [~, p] = size(X);   
    Ip = eye(p);
    XY = X'*Y;
    invIpXY = Ip/(Ip + XY);
%     XZ5 = invIpXY' - invIpXY;
%     XZ = 2 * XZ5;
    Z = 2 * (Y * invIpXY + X * (invIpXY' - Ip));
%     XZ = X' * Z;
    XZ = 2 * (invIpXY' - invIpXY);
    [TMP, ~, XXZ5] = Auxiliary(X, XZ, Z);
    VZ = horzcat(X, - Z + XXZ5);
    out = H + TMP * (VZ' * H);
end

% function out = VTbyCayley(X, Y, V)
%     % Y = retr(X, U);
%     U = inv_retr_cayley(X, Y);
%     [n, ~] = size(X);
%     In = eye(n);
% 
%     tmp = (U - 0.5 * X * (X' * U));
%     W = tmp * X';
%     W = 0.5 * (W - W');
%     Wl = 0.5 * [tmp, X]; 
%     Wr = [X, -tmp];
% %     out = (In - W) \ (In + W) * V;
%     out = woodbury_inverse(Wl, Wr) * (V + W*V);
% end

function [TMP, XZ5, XXZ5] = Auxiliary(X, XZ, Z)
    [~, p] = size(X);
    
    XZ5 = 0.5 * XZ;
    XXZ5 = X*XZ5;
    tmp1 = Z - XXZ5;
    UZ = horzcat(tmp1, X);
    tmp2 = Z'*Z - (0.75*XZ)' * XZ;
    tmp21 = horzcat(XZ5, eye(p));
    tmp22 = horzcat(tmp2, XZ5);
    I_VTU = 0.5 * vertcat(-tmp21, tmp22);
    % I_VTU = 0.5 * vertcat([-XZ5, -eye(p)], [tmp2, XZ5]);
    I_VTU = I_VTU + eye(2*p);
    tmp3 = inv(I_VTU);
    TMP = UZ * tmp3;

    % A = -XZ5 + eye(p);
    % B = eye(p);
    % C = tmp2;
    % D = XZ5 + eye(p);
    % invA = inv(A);
    % tmp4 = inv(-C * invA * B + D);
end

function out = woodbury_inverse(U,V)
    [n, p] = size(U);
    I_n = eye(n);
    I_p = eye(p);
    A = I_p - V'*U;
    out = I_n + (U / A) * V';
end


function out = projection(X, Y, U)
    YU = Y' * U;
    out = U - Y * (0.5 * (YU + YU'));
end

function out = retr_qr(X, U, a)
    if nargin == 3 && abs(a - 1) > 1e-16
        U = a * U;
    end
    [Q,R] = qr(X+U,0);
    tmp = sign(diag(R));
    out = Q.* tmp';
end

function out = inv_retr_qr(X, Y)    
    XY = X'*Y;
    H = 2*eye(size(X,2));
    R = solve_for_triu(XY, H);
    out = Y * R - X;
end