clear;close all;clc
rng(0);

set(0,'defaultaxesfontsize',22, ...
   'defaultaxeslinewidth',0.7, ...
   'defaultlinelinewidth',.8,'defaultpatchlinewidth',0.8);
set(0,'defaultlinemarkersize',10)


% load('School.mat')


r = 5;
m = 100;
Manifold = grassmannfactory(m, r);
Manifold.type = 'grassmann';
Manifold.transp = @isotransp;
lambda = .0;

nagents = 20;
ncostterms = 50;
K1 = 1;
%%% construct agents

prob = cell(nagents,1);
opts = cell(nagents,1);

U = Manifold.rand();

for i = 1:nagents   
    Z = {};
    for j = 1:ncostterms
        a = 10:1:50; 
        d = a(randi(40)); % d in [10, 50] 
        Xt = randn(d, m);
        wt = randn(m, 1);
        yt = Xt * (U * (U' * wt)) + 1e-6 * randn(d, 1);
        Z{j}.X = Xt;
        Z{j}.y = yt;
    end
    probtmp.M = Manifold;
    probtmp.ncostterms = ncostterms;
    probtmp = ConstructRFLLDMFLProb_Grass(Z, ncostterms, lambda, probtmp);
    prob{i} = probtmp;
                
    optstmp.batchsize = ceil(probtmp.ncostterms * 0.5);       
    optstmp.maxiter = K1;     
    opts{i} = optstmp;    
   
%     checkgradient(probtmp)
end


problem.M = Manifold;
problem.nagents = nagents;
problem.agent = prob;
problem.cost = @(x) cost(problem, x);
problem.grad = @(x) partialgrad(problem, x, 1:problem.nagents);

options.checkperiod = 1;
options.agent = opts;
options.maxiter = 100;
options.localMethod = 'RSGD';
options.option = 1;
options.verbosity = 1;

% for step size
% options.stepsize_type = 'decay';
options.stepsize_type = 'fix';
options.stepsize = .003;
options.s = nagents;

x0 = Manifold.rand();

Seed = 10;
%%%%%% K = 1 %%%%%%%
fvs1 = [];
ngfs1 = [];
dists1 = [];
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K1)
    [x1, fv1, info1] = RFedAGS(problem, x0, options);
    fvs1 = [fvs1, info1.fvs];
    ngfs1 = [ngfs1, info1.ngfs];
    tmp = [];
    for i = 1:length(info1.xs)
        tmp = [tmp; Manifold.dist(info1.xs{i}, U)];
    end
    dists1 = [dists1, tmp];
end
t = 0:options.checkperiod:options.maxiter;
figure(2) 
semilogy(t(2:end),mean(dists1,2), '-^', 'LineWidth',2)

%%%%%% K = 2 %%%%%%%
fvs2 = [];
ngfs2 = [];
dists2 = [];
K2 = 4;
for i = 1:nagents
    options.agent{i}.maxiter = K2;
end
for seed = 1:Seed
    fprintf('seed:%d, K:%d\n', seed, K2)
%     rng(seed)
    [x2, fv2, info2] = RFedAGS(problem, x0, options);
    fvs2 = [fvs2, info2.fvs];
    ngfs2 = [ngfs2, info2.ngfs];
    tmp = [];
    for i = 1:length(info2.xs)
        tmp = [tmp; Manifold.dist(info2.xs{i}, U)];
    end
    dists2 = [dists2, tmp];
end

figure(2)
hold on
semilogy(t(2:end),mean(dists2,2), '->', 'LineWidth',2)

%%%%%% K = 4 %%%%%%%
fvs3 = [];
ngfs3 = [];
dists3 = []; 
K3 = 8;
for i = 1:nagents
    options.agent{i}.maxiter = K3;
end
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K3)
    [x3, fv3, info3] = RFedAGS(problem, x0, options);
    fvs3 = [fvs3, info3.fvs];
    ngfs3 = [ngfs3, info3.ngfs];
    tmp = [];
    for i = 1:length(info3.xs)
        tmp = [tmp; Manifold.dist(info3.xs{i}, U)];
    end
    dists3 = [dists3, tmp];
end

figure(2)
hold on
semilogy(t(2:end),mean(dists3,2), '-v', 'LineWidth',2)

%%%%%% K = 8 %%%%%%%
fvs4 = [];
ngfs4 = [];
dists4 = [];
K4 = 10;
for i = 1:nagents
    options.agent{i}.maxiter = K4;
end
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K4)
    [x4, fv4, info4] = RFedAGS(problem, x0, options);
    fvs4 = [fvs4, info4.fvs];
    ngfs4 = [ngfs4, info4.ngfs];
    tmp = [];
    for i = 1:length(info4.xs)
        tmp = [tmp; Manifold.dist(info4.xs{i}, U)];
    end
    dists4 = [dists4, tmp];
end

figure(2)
hold on
semilogy(t(2:end),mean(dists4,2), '-<', 'LineWidth',2)
xlabel('Iteration','FontSize', 18)
ylabel('Distance to the optimal subspace','FontSize', 18)
legend('RFedAGS-K='+string(K1), 'RFedAGS-K='+string(K2), 'RFedAGS-K='+string(K3), 'RFedAGS-K='+string(K4), 'FontSize', 18)

saveas(gca, 'LDMFLGrass_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_Dist')

options.maxiter = 300;
[x5, fv5, info5] = rlbfgs(problem, x0, options);


figure(3)
semilogy(t, mean(fvs1,2)-fv5, '-^', t, mean(fvs2,2)-fv5, '->',...
         t, mean(fvs3,2)-fv5, '-v', t, mean(fvs4,2)-fv5, '-<', 'LineWidth', 2)
xlabel('Iteration','FontSize', 18)
ylabel('Excess Risk','FontSize', 18)
legend('RFedAGS-K='+string(K1), 'RFedAGS-K='+string(K2), 'RFedAGS-K='+string(K3), 'RFedAGS-K='+string(K4), 'FontSize', 18)

saveas(gca, 'LDMFLGrass_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms) + '_ER')

fvs1 = mean(fvs1,2);
dists1 = mean(dists1,2);
fvs2 = mean(fvs2,2);
dists2 = mean(dists2,2);
fvs3 = mean(fvs3,2);
dists3 = mean(dists3,2);
fvs4 = mean(fvs4,2);
dists4 = mean(dists4,2);

save('LDMFLGrass_Syn_K_Fixed'+string(nagents)+'x'+string(ncostterms)+'.mat',...
     "nagents", 'ncostterms',"fvs1","fvs2","fvs3","fvs4", "fv5",...
     "dists1", "dists2", "dists3", "dists4")

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function out = cost(problem, x)
    tmp = 0;
    for i = 1:problem.nagents
        tmp = tmp + problem.agent{i}.cost(x);
    end
    out = tmp / problem.nagents;
end

function out = partialgrad(problem, x, S)
    tmp = zeros(size(x));
    for i = reshape(S, 1, length(S))
        tmp = tmp + problem.agent{i}.grad(x);
    end
    out = tmp / length(S);
end


%%%%%%%% isometric vector transport by parallelization
function out = isotransp(x, y, v)    
    [n, p] = size(v);
    xbases = build_bases_embedded(x);
    ybases = build_bases_embedded(y);
    dim = p * (n - (p + 1) / 2);
    xbases_flat = (xbases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    ybases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    [U, ~, V] = svd(full(ybases_flat * xbases));
    v = reshape(v, n * p, 1);
    out = reshape(ybases * (U * (V' * (xbases_flat * v))), n, p);
end

function output = build_bases_embedded(x)
    [n, p] = size(x);
    x_perp = null(x');
    x_perp = x_perp(1 : n, 1 : n - p);   
    d = n * p - p * (p + 1) / 2;
    output = sparse(n * p, d);
    col = 1;
    for i = 2 : p
        x_i_1 = x(:, i - 1);
        for j = i : p
            output((i - 2) * n + 1 : (i - 1) * n, col) = - x(:, j);
            output((j - 1) * n + 1 : j * n, col) = x_i_1;
            col = col + 1;
        end
    end
    for i = 1 : p
        output((i - 1) * n + 1 : i * n, col : col + n - p - 1) = x_perp;
        col = col + n - p;
    end
end