clear;close all;clc
rng(0);

set(0,'defaultaxesfontsize',22, ...
   'defaultaxeslinewidth',0.7, ...
   'defaultlinelinewidth',.8,'defaultpatchlinewidth',0.8);
set(0,'defaultlinemarkersize',10)

% parkinson data
% [data1,data1_test] = generateTrainTestSplit('parkinson', 0.8);
% m = 19;
% nagents = 3;
% ncostterms = 14;

% school data
[data1,data1_test] = generateTrainTestSplit('school', 0.8);
m = 28;
nagents = 6;
ncostterms = 23;


%%% shuffle
per = randperm(length(data1));
data1 = data1(per);
data1_test = data1_test(per);

r = 5;
Manifold = grassmannfactory(m, r);
Manifold.type = 'grassmann';
% Manifold.isotransp = @Manifold.transp;
Manifold.transp = @isotransp;
lambda = -1.e-3;

% x0 = Manifold.rand();
% x1 = Manifold.rand();
% U1 = Manifold.randvec(x0);
% U2 = Manifold.randvec(x0);
% 
% U11 = Manifold.isotransp(x0, x1, U1);
% U22 = Manifold.isotransp(x0, x1, U2);
% 
% a = Manifold.inner(x0, U1, U2)
% b = Manifold.inner(x1, U11, U22)
% 
% a - b
% 
% U12 = Manifold.proj(x1, U11);
% norm(U11 - U12)


K1 = 1;
%%% construct agents

prob = cell(nagents,1);
opts = cell(nagents,1);

test = cell(nagents,1);
for i = 1:nagents   
    Z = {};
    Z_test = {};
    for j = 1:ncostterms
        Xt = data1((i-1)*ncostterms + j).X;
        yt = data1((i-1)*ncostterms + j).y;        
        Z{j}.X = Xt;
        Z{j}.y = yt;

        Xt_test = data1_test((i-1)*ncostterms + j).X;
        yt_test = data1_test((i-1)*ncostterms + j).y;        
        Z_test{j}.X = Xt_test;
        Z_test{j}.y = yt_test;        
    end
    test{i} = Z_test;

    probtmp.M = Manifold;
    probtmp.ncostterms = ncostterms;
    probtmp = ConstructRFLLDMFLProb_Grass(Z, ncostterms, lambda, probtmp);
    prob{i} = probtmp;        
        
    optstmp.batchsize = max(floor(probtmp.ncostterms * .8), 1);
    optstmp.checkperiod = 1;
    optstmp.maxiter = K1;     
    opts{i} = optstmp;    
   
%     checkgradient(probtmp)
end


problem.M = Manifold;
problem.nagents = nagents;
problem.agent = prob;
problem.cost = @(x) cost(problem, x);
problem.partialgrad = @(x, S) partialgrad(problem, x, S);
problem.grad = @(x) partialgrad(problem, x, 1:problem.nagents);
problem.ncostterms = nagents;
problem.s = nagents / 2;

problem.prediction = @(U, D) prediction(problem, U, D);
problem.NMSE = @(U, D) NMSE(problem, U, D, nagents);


options.checkperiod = 1;
options.agent = opts;
options.maxiter = 100;
options.localMethod = 'RSGD';
options.option = 1;
options.decaysteps = 5;
options.verbosity = 1;

% for step size
% options.stepsize_type = 'decay';
options.stepsize_type = 'fix';
options.stepsize = 1.e-6;
options.s = nagents / 2;

rng(0)
x0 = Manifold.rand();

Seed = 10;
%%%%%% K = 1 %%%%%%%

% RSGD
t = 0:options.checkperiod:options.maxiter;
fvs1_ = zeros(length(t), Seed);
ngfs1_ = zeros(length(t), Seed);
nmse1_ = zeros(length(t), Seed);
for seed = 1:Seed
%     rng(seed)
    fprintf('RSGD, seed:%d, K:%d\n', seed, K1)
    [x1_, fv1_, info1_] = RSGD(problem, x0, options);
    fvs1_(:, seed) = info1_.fvs;
    ngfs1_(:, seed) = info1_.ngfs;
    tmp = problem.NMSE(x0, test);
    for i = 1:length(info1_.xs)
        tmp = [tmp; problem.NMSE(info1_.xs{i}, test)];
    end
    nmse1_(:,seed) = tmp;
end


%-------
% RFedAGS
fvs1 = zeros(length(t), Seed);
ngfs1 = zeros(length(t), Seed);
nmse1 = zeros(length(t), Seed);
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K1)
    [x1, fv1, info1] = RFedAGS(problem, x0, options);
    fvs1(:, seed) = info1.fvs;
    ngfs1(:, seed) = info1.ngfs;
    tmp = problem.NMSE(x0, test);
    for i = 1:length(info1.xs)
        tmp = [tmp; problem.NMSE(info1.xs{i}, test)];
    end
    nmse1(:, seed) = tmp;
end


%%%%%% K = 4 %%%%%%%
fvs2 = [];
ngfs2 = [];
nmse2 = [];
K2 = 4;
for i = 1:nagents
    options.agent{i}.maxiter = K2;
end
for seed = 1:Seed
    fprintf('seed:%d, K:%d\n', seed, K2)
%     rng(seed)
    [x2, fv2, info2] = RFedAGS(problem, x0, options);
    fvs2 = [fvs2, info2.fvs];
    ngfs2 = [ngfs2, info2.ngfs];
    tmp = problem.NMSE(x0, test);
    for i = 1:length(info2.xs)
        tmp = [tmp; problem.NMSE(info2.xs{i}, test)];
    end
    nmse2 = [nmse2, tmp];
end

%%%%%% K = 8 %%%%%%%
fvs3 = [];
ngfs3 = [];
nmse3 = [];
K3 = 8;
for i = 1:nagents
    options.agent{i}.maxiter = K3;
end
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K3)
    [x3, fv3, info3] = RFedAGS(problem, x0, options);
    fvs3 = [fvs3, info3.fvs];
    ngfs3 = [ngfs3, info3.ngfs];
    tmp = problem.NMSE(x0, test);
    for i = 1:length(info3.xs)
        tmp = [tmp; problem.NMSE(info3.xs{i}, test)];
    end
    nmse3 = [nmse3, tmp];
end


%%%%%% K = 10 %%%%%%%
fvs4 = [];
ngfs4 = [];
nmse4 = [];
K4 = 10;
for i = 1:nagents
    options.agent{i}.maxiter = K4;
end
for seed = 1:Seed
%     rng(seed)
    fprintf('seed:%d, K:%d\n', seed, K4)
    [x4, fv4, info4] = RFedAGS(problem, x0, options);
    fvs4 = [fvs4, info4.fvs];
    ngfs4 = [ngfs4, info4.ngfs];
    tmp = problem.NMSE(x0, test);
    for i = 1:length(info4.xs)
        tmp = [tmp; problem.NMSE(info4.xs{i}, test)];
    end
    nmse4 = [nmse4, tmp];
end


[x5, fv5, info5] = steepestdescent(problem, x0, options);
nmse5 = problem.NMSE(x0, test);
xs = [info5.x];
for i = 1:options.maxiter
    nmse5(end+1) = problem.NMSE(xs{i}, test);
end


[x6, fv6, info6] = conjugategradient(problem, x0, options);
nmse6 = problem.NMSE(x0, test);
xs = [info6.x];
for i = 1:options.maxiter
    nmse6(end+1) = problem.NMSE(xs{i}, test);
end


[x7, fv7, info7] = rlbfgs(problem, x0, options);
nmse7 = problem.NMSE(x0, test);
xs = [info7.x];
for i = 1:options.maxiter
    nmse7(end+1) = problem.NMSE(xs{i}, test);
end

fvs1 = mean(fvs1, 2);
fvs1_ = mean(fvs1_, 2);
fvs2 = mean(fvs2, 2);
fvs3 = mean(fvs3, 2);
fvs4 = mean(fvs4, 2);
fvs5 = [info5.cost];
fvs6 = [info6.cost];
fvs7 = [info7.cost];

figure(1)
t = 0:options.checkperiod:options.maxiter;
semilogy(t,fvs1, '-^','LineWidth',2,'Color',[0 0.4470 0.7410])

figure(1)
hold on
semilogy(t,fvs2, '->', 'LineWidth',2, 'Color',[0.8500 0.3250 0.0980])

figure(1)
hold on
semilogy(t,fvs3, '-v', 'LineWidth',2,'Color',[0.9290 0.6940 0.1250])

figure(1)
hold on
semilogy(t,fvs4, '-<', 'LineWidth',2,'Color',[0.4940 0.1840 0.5560])

figure(1)
hold on
semilogy(t,fvs1_, '--p', 'LineWidth',2,'Color','blue')

figure(1)
hold on
semilogy(0:length([info5.cost])-1,[info5.cost], '-o','LineWidth',2, 'Color',[0.4660 0.6740 0.1880])

figure(1)
hold on
semilogy(0:length([info6.cost])-1,[info6.cost], '-s','LineWidth',2, 'Color',[0.3010 0.7450 0.9330])

figure(1)
hold on
semilogy(0:length([info7.cost])-1,[info7.cost], '-d', 'LineWidth',2,'Color',[0.6350 0.0780 0.1840])

xlabel('Iteration','FontSize', 18)
ylabel('Cost','FontSize', 18)
legend('RFedAGS-K='+string(K1), 'RFedAGS-K='+string(K2),...
       'RFedAGS-K='+string(K3), 'RFedAGS-K='+string(K4), 'RSGD', 'RSD', 'RCG', 'RLBFGS', 'FontSize', 18)

saveas(gca, 'LDMFLGrass_Sch_K_Fixed'+string(nagents)+'x'+string(ncostterms) + 'r=' + string(r) + '_Cost')

%%%%%%%%%

nmse1 = mean(nmse1,2);
nmse1_ = mean(nmse1_,2);
nmse2 = mean(nmse2,2);
nmse3 = mean(nmse3,2);
nmse4 = mean(nmse4,2);

figure(3)
semilogy(t,mean(nmse1,2), '-^','LineWidth',2,'Color',[0 0.4470 0.7410])

figure(3)
hold on
semilogy(t,mean(nmse2,2), '->', 'LineWidth',2, 'Color',[0.8500 0.3250 0.0980])

figure(3)
hold on
semilogy(t,mean(nmse3,2), '-v', 'LineWidth',2,'Color',[0.9290 0.6940 0.1250])

figure(3)
hold on
semilogy(t,mean(nmse4,2), '-<', 'LineWidth',2,'Color',[0.4940 0.1840 0.5560])

figure(3)
hold on
semilogy(t,mean(nmse1_,2), '--p', 'LineWidth',2,'Color', 'blue')

figure(3)
hold on
semilogy(0:length([info5.cost])-1, nmse5, '-o','LineWidth',2, 'Color',[0.4660 0.6740 0.1880])

figure(3)
hold on
semilogy(0:length([info6.cost])-1, nmse6, '-s','LineWidth',2, 'Color',[0.3010 0.7450 0.9330])

figure(3)
hold on
semilogy(0:length([info7.cost])-1, nmse7, '-d', 'LineWidth',2,'Color',[0.6350 0.0780 0.1840])

xlabel('Iteration','FontSize', 18)
ylabel('NMSE','FontSize', 18)
legend('RFedAGS-K='+string(K1), 'RFedAGS-K='+string(K2),...
       'RFedAGS-K='+string(K3), 'RFedAGS-K='+string(K4), 'RSGD', 'RSD', 'RCG', 'RLBFGS', 'FontSize', 18)

saveas(gca, 'LDMFLGrass_Sch_K_Fixed'+string(nagents)+'x'+string(ncostterms) + 'r=' + string(r) + '_ANMSE')


fileID = fopen('anmse_r=' + string(r) + '.txt', 'w');
fprintf(fileID, '%d & %d & %d & %d & %s & %s & %s & %s \n', K1, K2, K3, K4, 'RSGD', 'RSD', 'RCG', 'RLBFGS');
fprintf(fileID, '%.2e(%d) & %.2e(%d) & %.2e(%d) & %.2e(%d) & %.2e(%d) & %.2e(%d) & %.2e(%d) & %.2e(%d) & \n', ...
        min(mean(nmse1, 2)), find(mean(nmse1, 2) == min(min(mean(nmse1, 2)))) - 1,...
        min(mean(nmse2, 2)), find(mean(nmse2, 2) == min(min(mean(nmse2, 2)))) - 1,...
        min(mean(nmse3, 2)), find(mean(nmse3, 2) == min(min(mean(nmse3, 2)))) - 1,...
        min(mean(nmse4, 2)), find(mean(nmse4, 2) == min(min(mean(nmse4, 2)))) - 1,...
        min(mean(nmse1_, 2)), find(mean(nmse1_, 2) == min(min(mean(nmse1_, 2)))) - 1,...                       
        min(nmse5), find(nmse5 == min(nmse5)) - 1, ...
        min(nmse6), find(nmse6 == min(nmse6)) - 1, ...
        min(nmse7), find(nmse7 == min(nmse7)) - 1);

save('LDMFLGrass_Sch_K_Fixed'+string(nagents)+'x'+string(ncostterms) + 'r=' + string(r) + '.mat', ...
     "fvs1", "fvs1_", "fvs2", "fvs3", "fvs4", "fvs5", "fvs6", "fvs7", ...
     "nmse1","nmse1_","nmse2","nmse3","nmse4","nmse5","nmse6","nmse7")

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function out = cost(problem, x)
    tmp = 0;
    for i = 1:problem.nagents
        tmp = tmp + problem.agent{i}.cost(x);
    end
    out = tmp / problem.nagents;
end

function out = partialgrad(problem, x, S)
    tmp = zeros(size(x));
    for i = reshape(S, 1, length(S))
        tmp = tmp + problem.agent{i}.grad(x);
    end
    out = tmp / length(S);
end

function out = NMSE(problem, U, D, S)
    pred = problem.prediction(U, D);
    tmpyhat = [];
    tmpy    = [];
    for i = S
        N = length(D{i});
        for j = 1:N
            tmpy = [tmpy; D{i}{j}.y];
            tmpyhat = [tmpyhat; pred{i}{j}];
        end
    end
    out = mean((tmpy - tmpyhat).^2) / var(tmpy);
end

% function out = NMSE(problem, U, D, S)    
%     tmpm = 0;
%     tmpd = 0;
%     tmpy = [];
%     tmp  = 0;
%     for i = 1:S        
%         % NMSE over total tasks
%         [tmp_, tmpm_, tmpd_, tmpy_] = problem.agent{i}.NMSE(U, D{i}, problem.agent{i}.ncostterms);
%         tmpm = tmpm + tmpm_;
%         tmpd = tmpd + tmpd_;
%         tmpy = [tmpy; tmpy_];
% 
%         % NMSE over S agents
%         tmp = tmp + tmp_;
%     end
%     out = tmp / S;
% %     out = tmpm / tmpd / var(tmpy);
% end

function out = prediction(problem, U, D)
    S = length(D);
    out = cell(S, 1);
    for i = 1:S
        out{i} = problem.agent{i}.prediction(U, D{i});
    end    
end


%%%%%%%% isometric vector transport by parallelization
function out = isotransp(x, y, v)    
    [n, p] = size(v);
    xbases = build_bases_embedded(x);
    ybases = build_bases_embedded(y);
    dim = p * (n - (p + 1) / 2);
    xbases_flat = (xbases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    ybases_flat = (ybases * spdiags([ones(p * (p - 1) / 2, 1) / 2; ones((n - p)* p, 1)], 0, dim, dim))';
    [U, ~, V] = svd(full(ybases_flat * xbases));
    v = reshape(v, n * p, 1);
    out = reshape(ybases * (U * (V' * (xbases_flat * v))), n, p);
end

function output = build_bases_embedded(x)
    [n, p] = size(x);
    x_perp = null(x');
    x_perp = x_perp(1 : n, 1 : n - p);   
    d = n * p - p * (p + 1) / 2;
    output = sparse(n * p, d);
    col = 1;
    for i = 2 : p
        x_i_1 = x(:, i - 1);
        for j = i : p
            output((i - 2) * n + 1 : (i - 1) * n, col) = - x(:, j);
            output((j - 1) * n + 1 : j * n, col) = x_i_1;
            col = col + 1;
        end
    end
    for i = 1 : p
        output((i - 1) * n + 1 : i * n, col : col + n - p - 1) = x_perp;
        col = col + n - p;
    end
end








